module type S = sig

  type 'a m

  val return : 'a -> 'a m
  val bind : 'a m -> ('a -> 'b m) -> 'b m

  (** Non-deterministic choice. *)
  val orelse : 'a m -> 'a m -> 'a m

  (** Failure. *)
  val fail : 'a m

  (** [iter m f] runs [m],
    * calling [f] on each possible result. *)
  val iter : 'a m -> ('a -> unit) -> unit

end

module List : S = struct
  type 'a m = 'a list
  let return x = [x]
  let bind m n = List.flatten (List.map n m)
  let orelse m n = m @ n
  let fail = []
  let iter m f = List.iter f m
end

(** On supprime les listes: spécialisation pour l'itération. *)
module Iter : S = struct
  type 'a m = ('a -> unit) -> unit
  let iter m f = m f
  let return x = fun f -> f x
  let orelse m n f = m f ; n f
  let fail f = ()
  let bind m n f = m (fun x -> n x f)
end

(* Une version intermédiaire entre Iter et Cont. *)
module IterCPS : S = struct
  type 'a m = ('a -> (unit -> unit) -> unit) -> (unit -> unit) -> unit
  let iter m f = m (fun x k -> f x ; k ()) (fun () -> ())
  let return x f k = f x k
  let orelse m n f k = m f (fun () -> n f k)
  let fail f k = k ()
  let bind m n f k = m (fun x k -> n x f k) k
end

module Cont : S = struct
  type r
  type 'a fk = unit -> r
  type 'a sk = 'a -> 'a fk -> r
  type 'a m = 'a sk -> 'a fk -> r
  let return x sk fk = sk x fk
  let bind f g sk fk = f (fun x fk -> g x sk fk) fk
  let orelse f g sk fk = f sk (fun () -> g sk fk)
  let fail sk fk = fk ()
  let iter (m:'a m) (f:'a->unit) =
    let exception End in
    try
      ignore (m (fun x fk -> f x ; fk ()) (fun () -> raise End)) ;
      assert false
    with End -> ()
end

module Sugar (M:S) = struct
  include M
  let (let*) = bind
  let (>>=) = bind
  let (>>>) f g = let* _ = f in g
end
